unit Main;

{$ifdef fpc}
  {$mode Delphi}
{$endif}

interface

uses
  {$ifdef fpc}
    LConvEncoding,
    FileUtil,
    ComCtrls,
  {$else}
    WinTypes,
    WinProcs,
    Spin,
  {$endif}
  Classes,
  SysUtils,
  Forms,
  Controls,
  Graphics,
  Dialogs,
  ExtCtrls,
  StdCtrls,
  SortGrid,
  WolfUtil,
  Grids;

type
  {$ifdef fpc}
    TSpinButton         = TUpDown;
  {$endif}

  { TMainForm }

  TMainForm             = class(TForm)
    Panel               : TPanel;
    FromColumn          : TEdit;
    FromUpDown          : TSpinButton;
    CountUpDown         : TSpinButton;
    ToUpDown            : TSpinButton;
    ToColumn            : TEdit;
    Count               : TEdit;
    ColFormatString     : TEdit;
    ColAlignment        : TRadioGroup;
    ColFormatType       : TRadioGroup;
    ColSortOrder        : TRadioGroup;
    BtnAdd              : TButton;
    BtnClear            : TButton;
    BtnFill             : TButton;
    BtnInsert           : TButton;
    BtnDelete           : TButton;
    BtnSort             : TButton;
    cbSafe              : TCheckBox;
    cbHide              : TCheckBox;
    Label1              : TLabel;
    Label2              : TLabel;
    Label3              : TLabel;
    Label4              : TLabel;
    Label5              : TLabel;
    SortCols            : TLabel;
    Grid                : tSortGrid;
    procedure             BtnDeleteClick(Sender: TObject);
    procedure             BtnFillClick(Sender: TObject);
    procedure             BtnInsertClick(Sender: TObject);
    procedure             BtnSortClick(Sender: TObject);
    procedure             cbSafeChange(Sender: TObject);
    procedure             ColAlignmentClick(Sender: TObject);
    procedure             ColSortOrderClick(Sender: TObject);
    procedure             ColFormatStringChange(Sender: TObject);
    procedure             ColFormatTypeClick(Sender: TObject);
    procedure             CountChange(Sender: TObject);
    procedure             FromColumnChange(Sender: TObject);
    procedure             FormCreate(Sender: TObject);
    procedure             GridColRowDeleted(Sender: TObject; IsColumn: Boolean;
                                            sIndex, tIndex: LongInt);
    procedure             GridHeaderClick(Sender: TObject; IsColumn: Boolean;
                                          Index: LongInt);
    procedure             GridColRowMoved(Sender: TObject; IsColumn: Boolean;
                                          sIndex, tIndex: LongInt);
    procedure             GridColRowInserted(Sender: TObject; IsColumn: Boolean;
                                             sIndex, tIndex: LongInt);
    procedure             FormClose(Sender: TObject; var CloseAction: TCloseAction);
    procedure             GridMouseUp(Sender: TObject; Button: TMouseButton;
                                      Shift: TShiftState; X, Y: Integer);
    procedure             FromUpDownDownClick(Sender: TObject);
    procedure             FromUpDownUpClick(Sender: TObject);
    procedure             ToUpDownDownClick(Sender: TObject);
    procedure             ToUpDownUpClick(Sender: TObject);
    procedure             CountUpDownDownClick(Sender: TObject);
    procedure             CountUpDownUpClick(Sender: TObject);
    procedure             GridColumnMoved(Sender: TObject;
                                          FromIndex, ToIndex: LongInt);
    procedure             GridColRowExchanged(Sender: TObject; IsColumn: Boolean;
                                              sIndex, tIndex: LongInt);
    procedure             GridCompareCells(Sender: TObject;
                                           ACol, ARow, BCol, BRow : LongInt;
                                           var Result : integer);
    procedure             GridDrawCell(Sender: TObject; aCol, aRow: LongInt;
                                       aRect: TRect; aState: TGridDrawState);
    procedure             BtnClearClick(Sender: TObject);
    procedure             BtnAddClick(Sender: TObject);
    procedure             FormDestroy(Sender: TObject);
    procedure             ToColumnChange(Sender: TObject);
  private
    ReadColumn          : Boolean;
    CustomStrings       : tStringList;
    procedure             ColumnsChanged;
    procedure             EnableAll(IsEnabled : Boolean);
    procedure             Check(Edit : TEdit; cFrom, cTo : Integer);
    procedure             Up(Edit : tEdit);
    procedure             Down(Edit : tEdit);
  end;

var
  MainForm : TMainForm;

implementation

{$ifdef fpc}
  {$R *.lfm}
{$else}
  {$R *.dfm}
{$endif}

const
  SortCaption   = '&Sort';
  CancelCaption = '&Cancel';
  sCustom       = 'Custom';
  MaxCount      = 5;

type
  tDebugObject = class
    Pos        : String;
  end;

procedure TMainForm.FormCreate(Sender: TObject);
  var i : Integer;
  begin
    Grid.align       := alClient;
    ReadColumn 	     := False;
    BtnSort.Caption  := SortCaption;
    SortCols.Caption := '';
    CustomStrings    := tStringList.Create;
    CustomStrings.Add('least');
    CustomStrings.Add('less');
    CustomStrings.Add('few');
    CustomStrings.Add('more');
    CustomStrings.Add('most');
    for i := ord(Low(tAlignment)) to ord(High(tAlignment)) do
      case tAlignment(i) of
        taLeftJustify  : ColAlignment.Items.Add('taLeftJustify');
        taRightJustify : ColAlignment.Items.Add('taRightJustify');
        taCenter       : ColAlignment.Items.Add('taCenter');
      else
        ColAlignment.Items.Add('taUnknown');
      end;
    for i := ord(Low(tFormatType)) to ord(High(tFormatType)) do
      case tFormatType(i) of
        sgfString   : ColFormatType.Items.Add('sgfString');
        sgfText     : ColFormatType.Items.Add('sgfText');
        sgfInteger  : ColFormatType.Items.Add('sgfInteger');
        sgfFloat    : ColFormatType.Items.Add('sgfFloat');
        sgfDateTime : ColFormatType.Items.Add('sgfDateTime');
      else
        ColAlignment.Items.Add('sgfUnknown');
      end;
    for i := ord(Low(TSortOrder)) to ord(High(TSortOrder)) do
      case TSortOrder(i) of
        soAscending  : ColSortOrder.Items.Add('soAscending');
        soDescending : ColSortOrder.Items.Add('soDescending');
      else
        ColSortOrder.Items.Add('unknown');
      end;
    ColumnsChanged;
  end;

procedure TMainForm.GridColRowDeleted(Sender: TObject; IsColumn: Boolean;
                                      sIndex, tIndex: LongInt);
  begin
    if sIndex = 0 then;
    if tIndex = 0 then;
    if IsColumn then
      ColumnsChanged;
  end;

procedure TMainForm.GridColRowInserted(Sender: TObject; IsColumn: Boolean;
                                       sIndex, tIndex: LongInt);
  var     Col : Integer;
  begin
    if IsColumn then
      begin
        Col := StrToInt(FromColumn.Text);
        if Col >= sIndex then
          FromColumn.Text := IntToStr(Col + tIndex - sIndex + 1);
        ColumnsChanged;
      end;
  end;
  
procedure TMainForm.GridHeaderClick(Sender: TObject; IsColumn: Boolean;
                                    Index: LongInt);
  begin
    if IsColumn then
      begin
        FromColumn.Text := IntToStr(Index);
        ToColumn.Text   := FromColumn.Text;
      end;
  end;

procedure TMainForm.Up(Edit : tEdit);
  begin
    Edit.Text := IntToStr(StrToInt(Edit.Text) + 1);
  end;
  
procedure TMainForm.Down(Edit : tEdit);
  begin
    Edit.Text := IntToStr(StrToInt(Edit.Text) - 1);
  end;
  
procedure TMainForm.FromUpDownDownClick(Sender: TObject);
  begin
    Down(FromColumn);
  end;

procedure TMainForm.FromUpDownUpClick(Sender: TObject);
  begin
    Up(FromColumn);
  end;
  
procedure TMainForm.ToUpDownDownClick(Sender: TObject);
  begin
    Down(ToColumn);
  end;
  
procedure TMainForm.ToUpDownUpClick(Sender: TObject);
  begin
    Up(ToColumn);
  end;
  
procedure TMainForm.CountUpDownDownClick(Sender: TObject);
  begin
    Down(Count);
  end;
  
procedure TMainForm.CountUpDownUpClick(Sender: TObject);
  begin
    Up(Count);
  end;
  
procedure TMainForm.GridMouseUp(Sender: TObject; Button: TMouseButton;
                                Shift: TShiftState; X, Y: Integer);
  var     Col, Row : LongInt;
  begin
    if Button = mbLeft then;
    if Shift = [] then;
    Col := -1;
    Row := -1;
    Grid.MouseToCell(X, Y, Col, Row);
    if Row = 0 then
      begin
        FromColumn.Text := IntToStr(Col);
        ToColumn.Text   := FromColumn.Text;
      end;
  end;

procedure TMainForm.GridColRowMoved(Sender: TObject; IsColumn: Boolean;
                                    sIndex, tIndex: LongInt);
  var     Col : Integer;
  begin
    if IsColumn then
      begin
        Col := StrToInt(FromColumn.Text);
        if sIndex < tIndex then
          begin
            if (sIndex < Col) and (Col <= tIndex) then
              FromColumn.Text := IntToStr(Col - 1)
            else
              if (Col = tIndex) then
                FromColumn.Text := IntToStr(tIndex);
          end
        else
          if (tIndex <= Col) and (Col < sIndex) then
            FromColumn.Text := IntToStr(Col + 1)
          else
            if (Col = tIndex) then
              FromColumn.Text := IntToStr(tIndex);
        ReadColumn := true;
      end;
  end;

procedure TMainForm.GridColumnMoved(Sender: TObject;
                                    FromIndex, ToIndex: LongInt);
  var     Col : Integer;
  begin
    Col := StrToInt(FromColumn.Text);
    if FromIndex < ToIndex then
      begin
        if (FromIndex < Col) and (Col <= ToIndex) then
          FromColumn.Text := IntToStr(Col - 1)
        else
          if (Col = FromIndex) then
            FromColumn.Text := IntToStr(ToIndex);
      end
    else
      if (ToIndex <= Col) and (Col < FromIndex) then
        FromColumn.Text := IntToStr(Col + 1)
      else
        if (Col = FromIndex) then
          FromColumn.Text := IntToStr(ToIndex);
    ReadColumn := true;
  end;

procedure TMainForm.FormClose(Sender: TObject; var CloseAction: TCloseAction);
  begin
    if CloseAction = caFree then;
    Grid.MayContinue := False;
  end;

procedure TMainForm.BtnDeleteClick(Sender: TObject);
  begin
    Grid.DeleteCols(StrToInt(FromColumn.Text), StrToInt(ToColumn.Text));
    {$ifndef fpc}
      ColumnsChanged;
    {$endif}
  end;

procedure TMainForm.BtnInsertClick(Sender: TObject);
  var     Col,
          n    : Integer;
  begin
    Col := StrToInt(FromColumn.Text);
    n   := StrToInt(Count.Text);
    Grid.InsertCols(Col, n);
   {$ifndef fpc}
      FromColumn.Text := IntToStr(Col + n);
      ColumnsChanged;
   {$endif}
  end;

procedure TMainForm.GridColRowExchanged(Sender: TObject; IsColumn: Boolean;
                                        sIndex, tIndex: LongInt);
  begin
    if sIndex = 0 then;
    if tIndex = 0 then;
    if IsColumn then
      ColumnsChanged;
  end;

procedure TMainForm.GridCompareCells(Sender: TObject;
                                     ACol, ARow, BCol, BRow: LongInt;
                                     var Result: integer);
  begin
    if ACol <> BCol then
      Result := Grid.DefaultCompareCells(ACol, ARow, BCol, BRow)
    else
      if Grid.Cells[ACol, 0] <> sCustom then
        Result := Grid.DefaultCompareCells(ACol, ARow, BCol, BRow)
      else
        Result := CustomStrings.IndexOf(Grid.Cells[ACol, ARow])
                  - CustomStrings.IndexOf(Grid.Cells[BCol, BRow]);
  end;

procedure TMainForm.ColAlignmentClick(Sender: TObject);
  begin
    Grid.ColAlignment[StrToInt(FromColumn.Text)] :=
      tAlignment(ColAlignment.ItemIndex);
  end;

procedure TMainForm.EnableAll(IsEnabled : Boolean);
  begin
    if IsEnabled or cbSafe.Checked then
      begin
        FromUpDown.Enabled      := IsEnabled;
        CountUpDown.Enabled     := IsEnabled;
        ToUpDown.Enabled        := IsEnabled;
        FromColumn.Enabled      := IsEnabled;
        ToColumn.Enabled        := IsEnabled;
        Count.Enabled           := IsEnabled;
        ColFormatString.Enabled := IsEnabled;
        ColAlignment.Enabled    := IsEnabled;
        ColFormatType.Enabled   := IsEnabled;
        ColSortOrder.Enabled    := IsEnabled;
        BtnAdd.Enabled          := IsEnabled;
        BtnClear.Enabled        := IsEnabled;
        BtnFill.Enabled         := IsEnabled;
        BtnInsert.Enabled       := IsEnabled;
        BtnDelete.Enabled       := IsEnabled;
        Grid.Enabled            := IsEnabled;
      end;
    if IsEnabled then
      begin
        BtnSort.Caption := SortCaption;
        Screen.Cursor   := crDefault;
      end
    else
      begin
        BtnSort.Caption := CancelCaption;
        Screen.Cursor   := crHourGlass;
      end;
  end;

procedure TMainForm.BtnSortClick(Sender: TObject);
  begin
    if BtnSort.Caption = CancelCaption then
      Grid.MayContinue := False
    else
      begin
        EnableAll(False);
        if cbHide.Checked then
          Grid.Hide;
        Grid.SortRows(1, Grid.RowCount - 1);
        Grid.Show;
        EnableAll(True);
      end;
  end;

procedure TMainForm.Check(Edit : TEdit; cFrom, cTo : Integer);
  var     i : Integer;
  begin
    if cTo >= cFrom then
      begin
        i := StrToInt(Edit.Text);
        if i < cFrom then
          Edit.Text := IntToStr(cFrom)
        else
          if i > cTo then
            Edit.Text := IntToStr(cTo);
      end;
  end;

procedure TMainForm.cbSafeChange(Sender: TObject);
  begin
    if cbSafe.Checked then
      begin
        {$ifdef fpc}
          if Grid.ColCount > 0 then
            begin
              FromUpDown.Max := Grid.ColCount - 1;
              ToUpDown.Max   := Grid.ColCount - 1;
            end
          else
            begin
              FromUpDown.Max := 0;
              ToUpDown.Max   := 0;
            end;
          CountUpDown.Max    := MaxCount;
        {$else}
          Check(FromColumn, 0, Grid.ColCount - 1);
          Check(ToColumn,   0, Grid.ColCount - 1);
          Check(Count,      1, MaxCount);
        {$endif}
      end
    else
      begin
        {$ifdef fpc}
          FromUpDown.Max  := 100;
          ToUpDown.Max    := 100;
          CountUpDown.Max := 100;
        {$endif}
      end;
  end;

procedure TMainForm.BtnFillClick(Sender: TObject);
  const   ColCount     = ord(High(tFormatType)) + 2;
          RowCount     = 4000;
  var     Col, Row,
          DateTimeCol,
          CustomCol    : Integer;
          UsedTypes1,
          UsedTypes2   : set of tFormatType;
          t            : tFormatType;
  begin
    for Col := 0 to Grid.ColCount - 1 do
      for Row := 0 to Grid.RowCount - 1 do
        if Grid.Objects[Col, Row] <> Nil then
          Grid.Objects[Col, Row].Free;
    if Grid.ColCount < ColCount then
      Grid.ColCount := ColCount;
    if Grid.RowCount < RowCount then
      Grid.RowCount := RowCount;
    UsedTypes1 := [];
    CustomCol  := -1;
    for Col := 0 to Grid.ColCount - 1 do
      begin
        UsedTypes1 := UsedTypes1 + [Grid.ColFormatType[Col]];
        if Grid.Cells[Col, 0] = sCustom then
          CustomCol := Col;
      end;
    for t := Low(tFormatType) to High(tFormatType) do
      if not (t in UsedTypes1) then
        begin
          UsedTypes2 := [];
          for Col := 0 to Grid.ColCount - 1 do
            begin
              if (Grid.ColFormatType[Col] in UsedTypes2)
                 and (Col <> CustomCol)
              then
                Grid.ColFormatType[Col] := t;
              UsedTypes2 := UsedTypes2 + [Grid.ColFormatType[Col]];
            end;
        end;
    DateTimeCol := -1;
    UsedTypes1  := [];
    for Col := 0 to Grid.ColCount - 1 do
      begin
        if not (Grid.ColFormatType[Col] in UsedTypes1)
           and (Col <> CustomCol)
        then
          begin
            UsedTypes1 := UsedTypes1 + [Grid.ColFormatType[Col]];
            if Grid.ColFormatType[Col] = sgfDateTime then
              DateTimeCol := Col;
          end
        else
          if CustomCol < 0 then
            begin
              CustomCol := Col;
              Grid.ColFormatType[Col] := sgfText;
            end;
      end;
    if DateTimeCol >= 0 then
      if Grid.ColFormatString[DateTimeCol] = '' then
        Grid.ColFormatString[DateTimeCol] :=
          GetShortDateFormat + ' ' + GetLongTimeFormat;
    for Col := 0 to Grid.ColCount - 1 do
      begin
        if Col = CustomCol then
          begin
            Grid.ColAlignment[Col] := taLeftJustify;
            Grid.Cells[Col, 0]     := 'Custom';
          end
        else
          begin
            case Grid.ColFormatType[Col] of
              sgfInteger  : begin
                              Grid.ColAlignment[Col] := taRightJustify;
                              Grid.Cells[Col, 0]     := 'Integer';
                            end;
              sgfFloat    : begin
                              Grid.ColAlignment[Col] := taRightJustify;
                              Grid.Cells[Col, 0]     := 'Float';
                            end;
              sgfDateTime : begin
                              Grid.ColAlignment[Col] := taRightJustify;
                              Grid.Cells[Col, 0]     := 'DateTime';
                            end;
              sgfText     : begin
                              Grid.ColAlignment[Col] := taCenter;
                              Grid.Cells[Col, 0]     := 'Not case-sensitive';
                            end;
              sgfString   : begin
                              Grid.ColAlignment[Col] := taLeftJustify;
                              Grid.Cells[Col, 0]     := 'Case-sensitive';
                            end;
            end;
          end;
        Grid.MinWidthCell(Col, 0);
        for Row := 1 to Grid.RowCount - 1 do
          begin
            if Col = CustomCol then
              Grid.Cells[Col, Row] := CustomStrings[Random(CustomStrings.Count)]
            else
              case Grid.ColFormatType[Col] of
                sgfInteger :
                  if row mod 10 <> 1 then
                    Grid.Cells[Col, Row] := Grid.Cells[Col, Row - 1]
                  else
                    Grid.Cells[Col, Row] :=
                      IntToStr(LongestInt(Random(MaxInt))
                               shl ((SizeOf(LongestInt) - SizeOf(Integer)) * 8)
                               - (MaxLongestInt div 2));
                sgfFloat :
                  if (row mod 10 <> 5) and (Row <> 1) then
                    Grid.Cells[Col, Row] := Grid.Cells[Col, Row - 1]
                  else
                    Grid.Cells[Col, Row] := Format('%.2f', [Random * 1000 - 500]);
                sgfDateTime :
                  if (row mod 10 <> 7) and (Row <> 1) then
                    Grid.Cells[Col, Row] := Grid.Cells[Col, Row - 1]
                  else
                    Grid.Cells[Col, Row] :=
                      {$ifdef ver80}
                        {Delphi 1}
                        FormatDateTime(Grid.ColFormatString[Col],
                          Random * 1000 + 720000);
                      {$else}
                        FormatDateTime(Grid.ColFormatString[Col],
                          Random * 1000 + 30000);
                      {$endif}
              else
                Grid.Cells[Col, Row] := GetLongMonthName(Random(12) + 1);
                if Random(2) = 0 then
                  Grid.Cells[Col, Row] := LowerCase(Grid.Cells[Col, Row]);
              end;
            Grid.Objects[Col, Row] := tDebugObject.Create;
            if Grid.Objects[Col, Row] <> Nil then
              (Grid.Objects[Col, Row] as tDebugObject).Pos :=
                IntToStr(Col) + '-' + IntToStr(Row);
            Grid.MinWidthCell(Col, Row);
          end;
      end;
    ColumnsChanged;
  end;

procedure TMainForm.ColSortOrderClick(Sender: TObject);
  begin
    Grid.ColSortOrder[StrToInt(FromColumn.Text)] :=
      TSortOrder(ColSortOrder.ItemIndex);
  end;

procedure TMainForm.ColFormatStringChange(Sender: TObject);
  begin
    Grid.ColFormatString[StrToInt(FromColumn.Text)] := ColFormatString.Text;
  end;

procedure TMainForm.ColFormatTypeClick(Sender: TObject);
  begin
    Grid.ColFormatType[StrToInt(FromColumn.Text)] :=
      tFormatType(ColFormatType.ItemIndex);
  end;

procedure TMainForm.CountChange(Sender: TObject);
  begin
    if cbSafe.Checked then
      Check(Count, 1, MaxCount);
  end;

procedure TMainForm.FromColumnChange(Sender: TObject);
  begin
    if cbSafe.Checked then
      Check(FromColumn, 0, Grid.ColCount - 1);
    ColumnsChanged;
  end;

procedure TMainForm.GridDrawCell(Sender: TObject; aCol, aRow: LongInt;
                                 aRect: TRect; aState: TGridDrawState);
  begin
    if aCol       = 0 then;
    if aRow       = 0 then;
    if aRect.Left = 0  then;
    if aState     = [] then;
    if ReadColumn then
      begin
        ReadColumn := False;
        ColumnsChanged;
      end;
  end;

procedure TMainForm.BtnClearClick(Sender: TObject);
  begin
    Grid.ClearSortCols;
    SortCols.Caption := '';
  end;

procedure TMainForm.ColumnsChanged;
  var     s : String;
          i : Integer;
  begin
    i := StrToInt(FromColumn.Text);
    if (0 <= i) and (i < Grid.ColCount) or not cbSafe.Checked then
      begin
        ColAlignment.ItemIndex  := ord(Grid.ColAlignment[i]);
        ColFormatType.ItemIndex := ord(Grid.ColFormatType[i]);
        ColSortOrder.ItemIndex  := ord(Grid.ColSortOrder[i]);
        ColFormatString.Text    := Grid.ColFormatString[i];
      end;
    s := '';
    for i := 0 to Grid.SortColCount - 1 do
      s := s + ', ' + IntToStr(Grid.SortCols[i]);
    SortCols.Caption := Copy(s, 3, length(s));
    if cbSafe.Checked then
      begin
        {$ifdef fpc}
          if Grid.ColCount > 0 then
            begin
              FromUpDown.Max := Grid.ColCount - 1;
              ToUpDown.Max   := Grid.ColCount - 1;
            end
          else
            begin
              FromUpDown.Max := 0;
              ToUpDown.Max   := 0;
            end;
        {$else}
          Check(FromColumn, 0, Grid.ColCount - 1);
          Check(ToColumn,   0, Grid.ColCount - 1);
        {$endif}
      end;
  end;

procedure TMainForm.BtnAddClick(Sender: TObject);
  var     i, iFrom, iTo : Integer;
  begin
    iFrom := StrToInt(FromColumn.Text);
    iTo   := StrToInt(ToColumn.Text);
    if iFrom < iTo then
      for i := iFrom to iTo do
        Grid.AddSortCol(i)
    else
      for i := iFrom downto iTo do
        Grid.AddSortCol(i);
    ColumnsChanged;
  end;

procedure TMainForm.FormDestroy(Sender: TObject);
  begin
    CustomStrings.Free;
  end;

procedure TMainForm.ToColumnChange(Sender: TObject);
  begin
    if cbSafe.Checked then
      Check(ToColumn, 0, Grid.ColCount - 1);
  end;

end.

